module Algebra where

import Data.List
import Term
import Rule
import SMT

type Precedence = [String]

type Linear = ([Bool], Int)

type Interpretation = (String, Linear)

type Algebra = [Interpretation]

type Value = ([String], Int)

-- [("", a0), ("x1", a1), ("x2", a2)] stands for a0 + a1 x1 + a2 x2.
type Coefficient = [(String, Exp)]

-- Pretty printers

varX :: Int -> String
varX i = "x" ++ show (i + 1)

showLinear :: Linear -> String
showLinear (bs, n)
  | terms == [] = "0"
  | otherwise   = intercalate " + " terms
  where
    ss = [ varX i | (i, b) <- zip [0..] bs, b ]
    terms = if n == 0 then ss else show n : ss

showInterpretation :: Interpretation -> String
showInterpretation (f, (bs, n)) =
  f ++ "_N(" ++ 
  intercalate "," [ varX i | i <- [0 .. length bs - 1] ] ++ ") = " ++
  showLinear (bs, n)

showInterpretations :: Algebra -> String
showInterpretations a = unlines [ "  " ++ showInterpretation x | x <- a ]

showPrecedence :: Precedence -> String
showPrecedence p = "  " ++ intercalate " > " p

-- Valuations

addValue :: Value -> Value -> Value
addValue (xs, m) (ys, n) = (xs ++ ys, m + n)

sumValue :: [Value] -> Value
sumValue vs = foldr addValue ([], 0) vs

valuate :: Algebra -> Term -> Value
valuate _ (V x) = ([x], 0)
valuate a (F f ts)
  | Just (bs, n) <- lookup f a =
      sumValue (([], n) : [ valuate a t | (t, b) <- zip ts bs, b ])
  | otherwise = error ("valuate: " ++ f ++ " is not assigned")

-- Orders

countElem :: Eq a => a -> [a] -> Int
countElem _ [] = 0
countElem x (y : ys)
  | x == y    = countElem x ys + 1
  | otherwise = countElem x ys

subsetAsMultiset :: Eq a => [a] -> [a] -> Bool
subsetAsMultiset xs ys =
  all (\x -> countElem x xs <= countElem x ys) xs

gtValue :: Value -> Value -> Bool
gtValue (xs, m) (ys, n) = m > n  && subsetAsMultiset ys xs

geqValue :: Value -> Value -> Bool
geqValue (xs, m) (ys, n) = m >= n  && subsetAsMultiset ys xs

gtA :: Algebra -> Term -> Term -> Bool
gtA w s t = gtValue (valuate w s) (valuate w t)

geqA :: Algebra -> Term -> Term -> Bool
geqA w s t = geqValue (valuate w s) (valuate w t)

mark :: String -> String
mark f = f ++ "#"

markTerm :: Term -> Term
markTerm (F f ts) = F (mark f) ts
markTerm _        = error "markTerm"

labeled :: Algebra -> String -> Bool
labeled a f = lookup (mark f) a /= Nothing

gtPrec :: Precedence -> String -> String -> Bool
gtPrec [] _ _ = False
gtPrec (f : fs) g h
  | f == g    = elem h fs
  | otherwise = gtPrec fs g h 

-- s = f(s1,...,sm) >_B f(t1,...,tn) = t if 
--  - f > g, or 
--  - f is labeled and s# >_A t#
gtB :: Precedence -> Algebra -> Term -> Term -> Bool
gtB p a s@(F f _) t@(F g _) = 
  gtPrec p f g ||
  (f == g && labeled a f && gtA a (markTerm s) (markTerm t))
gtB _ _ _ _ = error "gtB"

-- s = f(s1,...,sm) >=_B f(t1,...,tn) = t if s# >_A t#
geqB :: Algebra -> Term -> Term -> Bool
geqB a s@(F f _) t@(F g _) =
  f == g &&
  (not (labeled a f) || (labeled a f && geqA a (markTerm s) (markTerm t)))
geqB _ _ _ = error "geqB"

gtLex :: Eq a => (a -> a -> Bool) -> [a] -> [a] -> Bool
gtLex _ [] []              = False
gtLex gt (x : xs) (y : ys) = gt x y || (x == y && gtLex gt xs ys)
gtLex _ _ _ = error "gtLex"


-- Encoding functions.

-- f_A(x_1, ..., x_n) = af_0 + af_1 x_1 + ... + af_n x_n

idA :: String -> Int -> String
idA f i = "a" ++ f ++ "_" ++ show i

varA :: String -> Exp
varA f = Var (idA f 0)

fvarA :: String -> Int -> Formula
fvarA f i = FVar (idA f i)

interpret :: Term -> Coefficient
interpret (V x)    = [(x, Val 1)]
interpret (F f ts) = 
  ("", varA f) : 
  [ (x, Ite (fvarA f i) e (Val 0))
  | (i, t) <- zip [1..] ts, 
    (x, e) <- interpret t ]

coefficient :: Coefficient -> String -> Exp
coefficient a x = Plus [ e | (y, e) <- a, x == y ]


--  s >=_A t

encodeGeqA :: Term -> Term -> Formula
encodeGeqA s t =
  conj [ Geq (coefficient a x) (coefficient b x) 
       | x <- "" : Rule.variables (s, t) ]
  where
    a = interpret s
    b = interpret t

-- s >_A t
encodeGtA :: Term -> Term -> Formula
encodeGtA s t =
  conj 
    ([ Gt (coefficient a "") (coefficient b "") ] ++
     [ Geq (coefficient a x) (coefficient b x) | x <- Rule.variables (s, t) ])
  where
    a = interpret s
    b = interpret t


varPrec :: String -> Exp
varPrec f = Var ("p" ++ f)

idLabeled :: String -> String
idLabeled f = "l" ++ f

encodeLabeled :: String -> Formula
encodeLabeled f = FVar ("l" ++ f)

encodeGtPrec :: String -> String -> Formula
encodeGtPrec f g = Gt (varPrec f) (varPrec g)

encodeGtB :: Term -> Term -> Formula
encodeGtB s@(F f _) t@(F g _)
  | f == g    = conj [encodeLabeled f, encodeGtA (markTerm s) (markTerm t)]
  | otherwise = encodeGtPrec f g
encodeGtB _ _ = error "encodeGtB"

encodeGeqB :: Term -> Term -> Formula
encodeGeqB s@(F f _) t@(F g _)
  | f == g    = 
    disj [neg (encodeLabeled f), 
          conj [encodeLabeled f, encodeGeqA (markTerm s) (markTerm t)]]
  | otherwise = bottom
encodeGeqB _ _ = error "encodeGeqB"


-- Decoding functions.

decodePrecedence :: Model -> Precedence
decodePrecedence model =
  [ f | (_, f) <- sort [ (-n, f) | ('p' : f, n) <- model ] ]

decodeInterpretation :: Model -> String -> Int -> Interpretation
decodeInterpretation model f n =
  (f, ([ evalFormula model (fvarA f i) | i <- [1..n] ], 
       evalExp model (varA f)))

decodeAlgebra :: Model -> Signature -> [Interpretation]
decodeAlgebra model sig =
  [ decodeInterpretation model f n | (f, n) <- sig ] ++
  [ decodeInterpretation model (mark f) n | (f, n) <- sig, elem (idLabeled f, 1) model ]
